from typing import List, Tuple

import sphobjinv as soi

from .data import Database, Link, Source


def define_importer(f):
    """Function that handles the details of importing a list of links.

    The idea is that this function calls :code:`f` with the reference given on the
    command line to get the list of links to import. It then handles the details of
    updating the database with these links.
    """

    # This outer function needs to handle the args given on the command line.
    def link_importer(filepath, uri):

        db = Database(filepath, create=True)

        source, links = f(uri)
        source.uri = f"{f.__name__}://{uri}"

        for l in links:
            source.links.append(l)

        Source.add(db, items=[source], commit=False)
        Link.add(db, items=links)

    return link_importer


def importer(f=None):
    """Decorator for defining importers."""

    if f is None:
        return define_importer

    return define_importer(f)


@importer
def sphinx(uri: str) -> Tuple[Source, List[Link]]:
    """Import links from a sphinx documentation site."""

    if uri.endswith("/"):
        uri = f"{uri}objects.inv"

    if not uri.endswith("objects.inv"):
        uri = f"{uri}/objects.inv"

    inv = soi.Inventory(url=uri)
    print(f"Found object index: {uri} with {inv.count} entries")

    prefix = uri.replace("objects.inv", "")
    source = Source(name=f"{inv.project} - {inv.version} | Sphinx Docs", prefix=prefix)

    links = []
    for item in inv.objects:
        name = item.dispname_expanded
        url = item.uri_expanded

        links.append(Link(name=name, url=url))

    return source, links



import pathlib
import webbrowser

from sqlalchemy import create_engine, Column, ForeignKey, Integer, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship

Base = declarative_base()


class DbOps:
    """Mixin class for database operations with ORM models."""

    @classmethod
    def add(cls, db, items=None, commit=True, **kwargs):
        """Add a link or collection of links to the given database."""

        session = db.session

        if items is None:
            dbitem = cls(**kwargs)
            session.add(dbitem)

        elif isinstance(items[0], dict):
            dbitems = [cls(**args) for args in items]
            session.add_all(dbitems)

        else:
            session.add_all(items)

        if commit:
            db.commit()

    @classmethod
    def get(cls, db, id):
        """Get a link with the given id from the database."""

        session = db.session
        item = session.query(cls).filter(cls.id == id).first()
        db.commit()

        return item

    @classmethod
    def search(cls, db, name=None, top=10):
        """Search the given database for links."""

        session = db.session
        filters = []

        if name is not None:
            filters.append(cls.name.ilike(f"%{name}%"))

        if len(filters) > 0:
            items = session.query(cls).filter(*filters)
        else:
            items = session.query(cls)

        db.commit()
        return items[:top]

    @classmethod
    def remove(cls, db, id):
        """Remove the given link id from the database."""


class Source(Base, DbOps):
    """Represents a source that a link was imported from."""

    __tablename__ = "sources"

    id = Column(Integer, primary_key=True)
    """The id of the source."""

    name = Column(Text, nullable=False)
    """The name of the source."""

    prefix = Column(Text, nullable=True)
    """The prefix that should be added to each link, if given."""

    uri = Column(Text, nullable=False)
    """The uri that was used when importing the source."""

    links = relationship("Link", backref="source")
    """Any links that were imported with this source."""

    def __eq__(self, other):

        if not isinstance(other, Source):
            return False

        return all(
            [
                self.id == other.id,
                self.name == other.name,
                self.prefix == other.prefix,
                self.uri == other.uri,
            ]
        )

    def __repr__(self):
        return f"Source<{self.name}, {self.uri}>"


class Link(Base, DbOps):
    """Represents an individual link."""

    __tablename__ = "links"

    id = Column(Integer, primary_key=True)
    """The id of the link."""

    name = Column(Text, nullable=False)
    """The name of the link."""

    url = Column(Text, nullable=False)
    """The url of the link."""

    source_id = Column(Integer, ForeignKey("sources.id"), nullable=True)
    """The id of the source the link was added with, if applicable"""

    def __eq__(self, other):

        if not isinstance(other, Link):
            return False

        return all(
            [self.id == other.id, self.name == other.name, self.url == other.url]
        )

    def __repr__(self):
        return f"{self.name} <{self.url}>"

    @property
    def url_expanded(self):
        """The full url, including the link's prefix if set."""

        if self.source and self.source.prefix:
            return f"{self.source.prefix}{self.url}"

        return self.url

    @classmethod
    def open(cls, db, link_id):
        """Open the link with the given id"""

        link = cls.get(db, link_id)
        url = link.url_expanded
        webbrowser.open(url)


class Database:
    """Manages connections to the database."""

    def __init__(self, filepath, create=False, verbose=False):
        """Parameters

        :param filepath: The path to the database
        :param create: Optional. If :code:`True` the database will be created if it
                       doesn't already exist.
        :param verbose: Optional. If :code:`True` enable sqlaclhemy's logging of SQL
                        commands
        """

        self.filepath = pathlib.Path(filepath)
        self.engine = create_engine("sqlite:///" + filepath, echo=verbose)
        self.new_session = sessionmaker(bind=self.engine)
        self._session = None

        if create:
            Source.__table__.create(bind=self.engine, checkfirst=True)
            Link.__table__.create(bind=self.engine, checkfirst=True)

    def commit(self):

        if self._session:
            self._session.commit()
            return

        raise RuntimeError("There is no session to commit!")

    @property
    def exists(self):
        """Determine if the database exists on disk."""
        return self.filepath.exists()

    @property
    def session(self):
        """Return the current session object."""

        if self._session is None:
            self._session = self.new_session()

        return self._session



from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import has_focus
from prompt_toolkit.keys import Keys
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import VSplit, HSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import Label, TextArea

from llyfrau.data import Database, Link

CURSOR = ">> "
SEPARATOR = " | "


class Column:
    def __init__(self, title: str):
        self.title = title
        self.col = FormattedTextControl(text=[], focusable=False)

    def width(self):

        text = self.col.text

        if isinstance(text, list) and len(text) == 0:
            lines = [""]

        if isinstance(text, list) and len(text) > 0:

            if isinstance(text[0], str):
                lines = text

            if isinstance(text[0], tuple):
                lines = [t[1] for t in text]

        if isinstance(text, str):
            lines = text.split("\n")

        maxlen = max([len(line) for line in lines])
        return max(maxlen, len(self.title))

    def clear(self):
        """Clear the column's contents"""
        self.col.text = []


class LinkTable:
    def __init__(self, filepath):
        self.db = Database(filepath)
        cursor = FormattedTextControl(
            focusable=True, text=[("", CURSOR)], show_cursor=False
        )
        sep = Window(width=len(SEPARATOR), char=SEPARATOR, style="class:line")

        self.ids = Column("ID")
        self.names = Column("Name")
        self.sources = Column("Source")
        self.urls = Column("URL")

        table_header = VSplit(
            [
                Label(text="", width=len(CURSOR)),
                sep,
                Label(self.ids.title, width=self.ids.width, style="bold"),
                sep,
                Label(self.names.title, width=self.names.width, style="bold"),
                sep,
                Label(self.sources.title, width=self.sources.width, style="bold"),
                sep,
                Label(self.urls.title, width=self.urls.width, style="bold"),
            ]
        )
        self.selection = Window(cursor, width=len(CURSOR), style="class:selector")
        table_body = VSplit(
            [
                self.selection,
                sep,
                Window(self.ids.col, width=self.ids.width),
                sep,
                Window(self.names.col, width=self.names.width),
                sep,
                Window(self.sources.col, width=self.sources.width),
                sep,
                Window(self.urls.col, width=self.urls.width),
            ]
        )

        self.prompt = TextArea(
            multiline=False,
            focusable=True,
            accept_handler=self._do_search,
            get_line_prefix=self._get_prompt,
        )

        table = HSplit([table_header, table_body])
        layout = HSplit([table, self.prompt])

        kb = KeyBindings()

        @kb.add("c-c", eager=True)
        @kb.add("q", filter=has_focus(self.selection))
        def close(event):
            event.app.exit()

        @kb.add("s", filter=has_focus(self.selection))
        def start_search(event):
            event.app.layout.focus(self.prompt)

        @kb.add("o", filter=has_focus(self.selection))
        @kb.add(Keys.Enter, filter=has_focus(self.selection))
        def open_link(event):
            cursor = self.selection.content.text
            idx = len(cursor)

            selected = self.ids.col.text[idx - 1]
            link_id = int(selected[1])

            Link.open(self.db, link_id)

        @kb.add(Keys.Down, filter=has_focus(self.selection))
        def next_item(event):
            cursor = self.selection.content.text
            idx = len(cursor)
            max_idx = len(self.ids.col.text)

            if idx + 1 > max_idx:
                return

            cursor.insert(0, ("", "\n"))

        @kb.add(Keys.Up, filter=has_focus(self.selection))
        def prev_item(event):
            cursor = self.selection.content.text
            idx = len(cursor)

            if idx == 1:
                return

            cursor.pop(0)

        self.app = Application(layout=Layout(layout), key_bindings=kb)

    def run(self):
        self._do_search()
        self.app.run()

    def _do_search(self, inpt: Buffer = None):
        self.selection.content.text = [("", CURSOR)]
        self.ids.clear()
        self.names.clear()
        self.sources.clear()
        self.urls.clear()

        name = None

        if inpt is not None and len(inpt.text) > 0:
            name = inpt.text

        links = Link.search(self.db, name=name, top=10)

        for idx, link in enumerate(links):

            newline = "\n" if idx < len(links) - 1 else ""

            self.ids.col.text.append(("", f"{link.id}{newline}"))
            self.names.col.text.append(("", f"{link.name}{newline}"))
            self.urls.col.text.append(("", f"{link.url_expanded}{newline}"))

            source = "" if not link.source else link.source.name
            self.sources.col.text.append(("", f"{source}{newline}"))

        self.app.layout.focus(self.selection)

    def _get_prompt(self, line_no, other):

        if has_focus(self.prompt)():
            return "Search: "

        return "[s]earch | [o]pen | [q]uit"



import argparse
import inspect
import pathlib
import shutil
import sys

import pkg_resources

from llyfrau._version import __version__
from llyfrau.data import Database, Link, Source
from .tui import LinkTable


def add_link(filepath, url, name):
    db = Database(filepath, create=True)
    Link.add(db, name=name, url=url)


def find_sources(filepath):

    path = pathlib.Path(filepath)

    if not path.exists():
        print(f"Unable to find links database: {filepath}", file=sys.stderr)
        return -1

    ids = ["ID"]
    names = ["Name"]
    uris = ["URI"]
    prefixes = ["Prefix"]

    db = Database(filepath)

    for source in Source.search(db):
        ids.append(source.id)
        names.append(source.name)
        uris.append(source.uri)
        prefixes.append(source.prefix)

    print(format_table([ids, names, uris, prefixes]))


def open_link_ui(filepath):

    table_ui = LinkTable(filepath)
    table_ui.run()


def call_command(cmd, args):

    params = inspect.signature(cmd).parameters
    cmd_args = {name: getattr(args, name) for name in params}
    return cmd(**cmd_args)


def format_cell(text, width, placeholder=None):

    if placeholder is None:
        placeholder = "..."

    if len(text) <= width:
        return text.ljust(width)

    idx = width - len(placeholder)
    return text[:idx] + placeholder


def format_column(col):
    return ["" if c is None else str(c) for c in col]


def format_table(columns):

    term_width, _ = shutil.get_terminal_size()
    maxwidth = (term_width - len(columns) + 1) // len(columns)

    columns = [format_column(col) for col in columns]
    sizes = [max(len(c) for c in col) for col in columns]
    widths = [min(maxwidth, size) for size in sizes]

    fmt = " ".join(["{:<" + str(n) + "}" for n in widths])
    rows = []

    for i in range(len(columns[0])):
        row = [
            format_cell(col[i], width=widths[idx]) for idx, col in enumerate(columns)
        ]
        rows.append(fmt.format(*row))

    return "\n".join(rows)


def _load_importers(parent):
    """Load importers and attach them to the cli interface."""

    for imp in pkg_resources.iter_entry_points("llyfrau.importers"):
        cmd = parent.add_parser(imp.name, help=f"{imp.name} importer")
        cmd.add_argument("uri", help="where to import links from")

        impl = imp.load()
        cmd.set_defaults(run=impl)


cli = argparse.ArgumentParser()
cli.add_argument(
    "-f",
    "--filepath",
    type=str,
    help="filepath to the links database",
    default="links.db",
)
cli.add_argument("--version", action="store_true", help="show version and exit")

commands = cli.add_subparsers(title="commands")
add = commands.add_parser("add", help="add a link")
add.add_argument("url", help="the link to add")
add.add_argument("-n", "--name", help="name of the link")
add.set_defaults(run=add_link)

import_ = commands.add_parser("import", help="import links from a source")
importers = import_.add_subparsers(title="importers")
_load_importers(importers)

sources = commands.add_parser("sources", help="list all link sources")
sources.set_defaults(run=find_sources)

open_ = commands.add_parser("open", help="open a link")
open_.set_defaults(run=open_link_ui)


def main():

    args = cli.parse_args()

    if args.version:
        print(f"llyfr v{__version__}")
        return 0

    if hasattr(args, "run"):
        return call_command(args.run, args)

    cli.print_help()



import os

from setuptools import find_packages, setup

info = {}
version = os.path.join("llyfrau", "_version.py")

with open(version) as f:
    exec(f.read(), info)


def readme():
    with open("README.md") as f:
        return f.read()


install_requires = ["prompt_toolkit", "sphobjinv", "sqlalchemy"]
extras = {"dev": ["black", "flake8", "pytest", "pytest-cov", "tox",]}

setup(
    name="llyfrau",
    version=info["__version__"],
    project_urls={
        "Source": "https://github.com/alcarney/llyfrau",
        "Tracker": "https://github.com/alcarney/llyfrau/issues",
    },
    description="Bookmark management",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Alex Carney",
    author_email="alcarneyme@gmail.com",
    license="MIT",
    packages=find_packages(".", exclude=["tests"]),
    python_requires=">=3.6",
    install_requires=install_requires,
    extras_require=extras,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={
        "console_scripts": ["llyfr = llyfrau.__main__:main"],
        "llyfrau.importers": ["sphinx = llyfrau.importers:sphinx"],
    },
)
